// Generated by make_intl_data.py. DO NOT EDIT.

#ifndef intl_components_MeasureUnitGenerated_h
#define intl_components_MeasureUnitGenerated_h

namespace mozilla::intl {

struct SimpleMeasureUnit {
  const char* const type;
  const char* const name;
};

/**
 * The list of currently supported simple unit identifiers.
 *
 * The list must be kept in alphabetical order of |name|.
 */
inline constexpr SimpleMeasureUnit simpleMeasureUnits[] = {
    // clang-format off
  {"area", "acre"},
  {"digital", "bit"},
  {"digital", "byte"},
  {"temperature", "celsius"},
  {"length", "centimeter"},
  {"duration", "day"},
  {"angle", "degree"},
  {"temperature", "fahrenheit"},
  {"volume", "fluid-ounce"},
  {"length", "foot"},
  {"volume", "gallon"},
  {"digital", "gigabit"},
  {"digital", "gigabyte"},
  {"mass", "gram"},
  {"area", "hectare"},
  {"duration", "hour"},
  {"length", "inch"},
  {"digital", "kilobit"},
  {"digital", "kilobyte"},
  {"mass", "kilogram"},
  {"length", "kilometer"},
  {"volume", "liter"},
  {"digital", "megabit"},
  {"digital", "megabyte"},
  {"length", "meter"},
  {"duration", "microsecond"},
  {"length", "mile"},
  {"length", "mile-scandinavian"},
  {"volume", "milliliter"},
  {"length", "millimeter"},
  {"duration", "millisecond"},
  {"duration", "minute"},
  {"duration", "month"},
  {"duration", "nanosecond"},
  {"mass", "ounce"},
  {"concentr", "percent"},
  {"digital", "petabyte"},
  {"mass", "pound"},
  {"duration", "second"},
  {"mass", "stone"},
  {"digital", "terabit"},
  {"digital", "terabyte"},
  {"duration", "week"},
  {"length", "yard"},
  {"duration", "year"},
    // clang-format on
};

}  // namespace mozilla::intl

#endif
